@@ -8,6 +8,7 @@ import sys
 import threading
 from typing import Dict, Optional
 
+import xmltodict
 from requests.models import Response
 
 from localstack import config
@@ -21,7 +22,10 @@ from localstack.constants import (
     LS_LOG_TRACE_INTERNAL,
     PATH_USER_REQUEST,
 )
-from localstack.services.cloudwatch.cloudwatch_listener import PATH_GET_RAW_METRICS
+from localstack.http import Router
+from localstack.http.dispatcher import Handler, handler_dispatcher
+from localstack.http.router import RegexConverter
+from localstack.runtime import events
 from localstack.services.generic_proxy import ProxyListener, modify_and_forward, start_proxy_server
 from localstack.services.infra import PROXY_LISTENERS
 from localstack.services.plugins import SERVICE_PLUGINS
@@ -31,18 +35,16 @@ from localstack.utils import persistence
 from localstack.utils.aws import aws_stack
 from localstack.utils.aws.aws_stack import is_internal_call_context, set_default_region_in_headers
 from localstack.utils.aws.request_routing import extract_version_and_action, matches_service_action
-from localstack.utils.common import (
-    TMP_THREADS,
-    empty_context_manager,
-    get_service_protocol,
-    is_port_open,
-    is_root,
-    parse_request_data,
-    run,
-)
-from localstack.utils.common import safe_requests as requests
-from localstack.utils.common import sleep_forever, start_thread, to_bytes, to_str, truncate
+from localstack.utils.functions import empty_context_manager
+from localstack.utils.http import parse_request_data
+from localstack.utils.http import safe_requests as requests
+from localstack.utils.net import is_port_open
+from localstack.utils.run import is_root, run
 from localstack.utils.server.http2_server import HTTPErrorResponse
+from localstack.utils.strings import to_bytes, to_str, truncate
+from localstack.utils.sync import sleep_forever
+from localstack.utils.threads import TMP_THREADS, start_thread
+from localstack.utils.xml import strip_xmlns
 
 LOG = logging.getLogger(__name__)
 
@@ -65,6 +67,7 @@ API_UNKNOWN = "_unknown_"
 # APIs for which no gzip encoding should be applied when returning the response
 HEADER_SKIP_RESPONSE_ZIPPING = "_skip_response_gzipping_"
 SKIP_GZIP_APIS = [S3]
+S3CONTROL_COMMON_PATH = "/v20180820/"
 
 
 class ProxyListenerEdge(ProxyListener):
@@ -73,15 +76,17 @@ class ProxyListenerEdge(ProxyListener):
         self.service_manager = service_manager or SERVICE_PLUGINS
 
     def forward_request(self, method, path, data, headers):
+        # kill the process if we receive this header
+        headers.get(HEADER_KILL_SIGNAL) and sys.exit(0)
+
+        if events.infra_stopping.is_set():
+            return 503
 
         if config.EDGE_FORWARD_URL:
             return do_forward_request_network(
                 0, method, path, data, headers, target_url=config.EDGE_FORWARD_URL
             )
 
-        # kill the process if we receive this header
-        headers.get(HEADER_KILL_SIGNAL) and sys.exit(0)
-
         target = headers.get("x-amz-target", "")
         auth_header = get_auth_string(method, path, headers, data)
         if auth_header and not headers.get("authorization"):
@@ -202,6 +207,16 @@ class ProxyListenerEdge(ProxyListener):
                     response.content,
                 )
 
+        # convert XML responses to JSON - TODO: introduce generalizable response conversion
+        if headers.get("Accept") == "application/json":
+            try:
+                # simple heuristic to detect XML responses
+                if str(response._content or "").startswith("<"):
+                    content = xmltodict.parse(to_str(response._content))
+                    response._content = strip_xmlns(content)
+            except Exception as e:
+                LOG.debug("Unable to convert XML response to JSON", exc_info=e)
+
         if (
             response._content
             and headers.get("Accept-Encoding") == "gzip"
@@ -248,7 +263,7 @@ def do_forward_request_inmem(api, method, path, data, headers, port=None):
     # TODO determine client address..?
     client_address = LOCALHOST_IP
     server_address = headers.get("host") or LOCALHOST
-    forward_url = "http://%s:%s" % (LOCALHOST, backend_port)
+    forward_url = f"http://{LOCALHOST}:{backend_port}"
     response = modify_and_forward(
         method=method,
         path=path,
@@ -264,7 +279,7 @@ def do_forward_request_inmem(api, method, path, data, headers, port=None):
 
 def do_forward_request_network(port, method, path, data, headers, target_url=None):
     # TODO: enable per-service endpoints, to allow deploying in distributed settings
-    target_url = target_url or "%s://%s:%s" % (get_service_protocol(), LOCALHOST, port)
+    target_url = target_url or "%s://%s:%s" % (config.get_protocol(), LOCALHOST, port)
     url = "%s%s" % (target_url, path)
     response = requests.request(
         method, url, data=data, headers=headers, verify=False, stream=True, allow_redirects=False
@@ -315,6 +330,7 @@ def get_auth_string(method, path, headers, data=None):
     return ""
 
 
+# TODO: refactor this function -> returning the port is redundant (given the returned service name)
 def get_api_from_headers(headers, method=None, path=None, data=None):
     """Determine API and backend port based on "Authorization" or "Host" headers."""
 
@@ -343,41 +359,43 @@ def get_api_from_headers(headers, method=None, path=None, data=None):
     # Fallback rules and route customizations applied below
     if host.endswith("cloudfront.net"):
         path = path or "/"
-        result = "cloudfront", config.PORT_CLOUDFRONT
+        result = "cloudfront", config.service_port("cloudfront")
     elif target.startswith("AWSCognitoIdentityProviderService") or "cognito-idp." in host:
-        result = "cognito-idp", config.PORT_COGNITO_IDP
+        result = "cognito-idp", config.service_port("cognito-idp")
     elif target.startswith("AWSCognitoIdentityService") or "cognito-identity." in host:
-        result = "cognito-identity", config.PORT_COGNITO_IDENTITY
+        result = "cognito-identity", config.service_port("cognito-identity")
+    elif result[0] == "s3" and path.startswith(S3CONTROL_COMMON_PATH):
+        result = "s3control", config.service_port("s3control")
     elif result[0] == "s3" or uses_host_addressing(headers):
-        result = "s3", config.PORT_S3
+        result = "s3", config.service_port("s3")
     elif result[0] == "states" in auth_header or host.startswith("states."):
-        result = "stepfunctions", config.PORT_STEPFUNCTIONS
+        result = "stepfunctions", config.service_port("stepfunctions")
     elif "route53." in host:
-        result = "route53", config.PORT_ROUTE53
+        result = "route53", config.service_port("route53")
     elif result[0] == "monitoring":
-        result = "cloudwatch", config.PORT_CLOUDWATCH
+        result = "cloudwatch", config.service_port("cloudwatch")
     elif result[0] == "email":
-        result = "ses", config.PORT_SES
+        result = "ses", config.service_port("ses")
     elif result[0] == "execute-api" or ".execute-api." in host:
-        result = "apigateway", config.PORT_APIGATEWAY
+        result = "apigateway", config.service_port("apigateway")
     elif target.startswith("Firehose_"):
-        result = "firehose", config.PORT_FIREHOSE
+        result = "firehose", config.service_port("firehose")
     elif target.startswith("DynamoDB_"):
-        result = "dynamodb", config.PORT_DYNAMODB
+        result = "dynamodb", config.service_port("dynamodb")
     elif target.startswith("DynamoDBStreams") or host.startswith("streams.dynamodb."):
         # Note: DDB streams requests use ../dynamodb/.. auth header, hence we also need to update result_before
-        result = result_before = "dynamodbstreams", config.PORT_DYNAMODBSTREAMS
+        result = result_before = "dynamodbstreams", config.service_port("dynamodbstreams")
     elif result[0] == "EventBridge" or target.startswith("AWSEvents"):
-        result = "events", config.PORT_EVENTS
+        result = "events", config.service_port("events")
     elif target.startswith("ResourceGroupsTaggingAPI_"):
-        result = "resourcegroupstaggingapi", config.PORT_RESOURCEGROUPSTAGGINGAPI
+        result = "resourcegroupstaggingapi", config.service_port("resourcegroupstaggingapi")
     elif result[0] == "resource-groups":
-        result = "resource-groups", config.PORT_RESOURCE_GROUPS
+        result = "resource-groups", config.service_port("resource-groups")
     elif result[0] == "es" and path is not None and not path.startswith("/2015-01-01/"):
         # For OpenSearch, the auth header points to the API ("es").
         # However, if the path does _not_ start with /2015-01-01 (the API version path prefix for the only ES API
         # version) it is a request to the opensearch API.
-        result = "opensearch", config.PORT_OPENSEARCH
+        result = "opensearch", config.service_port("opensearch")
 
     return result[0], result_before[1] or result[1], path, host
 
@@ -403,24 +421,25 @@ def is_s3_form_data(data_bytes):
     return False
 
 
+# TODO: refactor this function -> returning the port is redundant (given the returned service name)
 def get_api_from_custom_rules(method, path, data, headers):
     """Determine backend port based on custom rules."""
 
     # API Gateway invocation URLs
     if ("/%s/" % PATH_USER_REQUEST) in path:
-        return "apigateway", config.PORT_APIGATEWAY
+        return "apigateway", config.service_port("apigateway")
 
     # detect S3 presigned URLs
     if "AWSAccessKeyId=" in path or "Signature=" in path:
-        return "s3", config.PORT_S3
+        return "s3", config.service_port("s3")
 
     # heuristic for SQS queue URLs
     if is_sqs_queue_url(path):
-        return "sqs", config.PORT_SQS
+        return "sqs", config.service_port("sqs")
 
     # DynamoDB shell URLs
     if path.startswith("/shell") or path.startswith("/dynamodb/shell"):
-        return "dynamodb", config.PORT_DYNAMODB
+        return "dynamodb", config.service_port("dynamodb")
 
     data_bytes = to_bytes(data or "")
     version, action = extract_version_and_action(path, data_bytes)
@@ -429,52 +448,51 @@ def get_api_from_custom_rules(method, path, data, headers):
         return to_str(search_str) in path or to_bytes(search_str) in data_bytes
 
     if path == "/" and b"QueueName=" in data_bytes:
-        return "sqs", config.PORT_SQS
+        return "sqs", config.service_port("sqs")
 
     if "Action=ConfirmSubscription" in path:
-        return "sns", config.PORT_SNS
+        return "sns", config.service_port("sns")
 
     if path.startswith("/2015-03-31/functions/"):
-        return "lambda", config.PORT_LAMBDA
+        return "lambda", config.service_port("lambda")
 
     if _in_path_or_payload("Action=AssumeRoleWithWebIdentity"):
-        return "sts", config.PORT_STS
+        return "sts", config.service_port("sts")
 
     if _in_path_or_payload("Action=AssumeRoleWithSAML"):
-        return "sts", config.PORT_STS
+        return "sts", config.service_port("sts")
 
-    # CloudWatch backdoor API to retrieve raw metrics
-    if path.startswith(PATH_GET_RAW_METRICS):
-        return "cloudwatch", config.PORT_CLOUDWATCH
+    if _in_path_or_payload("Action=AssumeRole"):
+        return "sts", config.service_port("sts")
 
     # SQS queue requests
     if _in_path_or_payload("QueueUrl=") and _in_path_or_payload("Action="):
-        return "sqs", config.PORT_SQS
+        return "sqs", config.service_port("sqs")
     if matches_service_action("sqs", action, version=version):
-        return "sqs", config.PORT_SQS
+        return "sqs", config.service_port("sqs")
 
     # SNS topic requests
     if matches_service_action("sns", action, version=version):
-        return "sns", config.PORT_SNS
+        return "sns", config.service_port("sns")
 
     # TODO: move S3 public URLs to a separate port/endpoint, OR check ACLs here first
     stripped = path.strip("/")
     if method in ["GET", "HEAD"] and stripped:
         # assume that this is an S3 GET request with URL path `/<bucket>/<key ...>`
-        return "s3", config.PORT_S3
+        return "s3", config.service_port("s3")
 
     # detect S3 URLs
     if stripped and "/" not in stripped:
         if method == "PUT":
             # assume that this is an S3 PUT bucket request with URL path `/<bucket>`
-            return "s3", config.PORT_S3
+            return "s3", config.service_port("s3")
         if method == "POST" and is_s3_form_data(data_bytes):
             # assume that this is an S3 POST request with form parameters or multipart form in the body
-            return "s3", config.PORT_S3
+            return "s3", config.service_port("s3")
 
     # detect S3 requests sent from aws-cli using --no-sign-request option
     if "aws-cli/" in headers.get("User-Agent", ""):
-        return "s3", config.PORT_S3
+        return "s3", config.service_port("s3")
 
     # S3 delete object requests
     if (
@@ -483,23 +501,23 @@ def get_api_from_custom_rules(method, path, data, headers):
         and b"<Delete" in data_bytes
         and b"<Key>" in data_bytes
     ):
-        return "s3", config.PORT_S3
+        return "s3", config.service_port("s3")
 
     # Put Object API can have multiple keys
     if stripped.count("/") >= 1 and method == "PUT":
         # assume that this is an S3 PUT bucket object request with URL path `/<bucket>/object`
         # or `/<bucket>/object/object1/+`
-        return "s3", config.PORT_S3
+        return "s3", config.service_port("s3")
 
     auth_header = headers.get("Authorization") or ""
 
     # detect S3 requests with "AWS id:key" Auth headers
     if auth_header.startswith("AWS "):
-        return "s3", config.PORT_S3
+        return "s3", config.service_port("s3")
 
     # certain EC2 requests from Java SDK contain no Auth headers (issue #3805)
     if b"Version=2016-11-15" in data_bytes:
-        return "ec2", config.PORT_EC2
+        return "ec2", config.service_port("ec2")
 
 
 def get_service_port_for_account(service, headers):
@@ -508,6 +526,11 @@ def get_service_port_for_account(service, headers):
 
 
 PROXY_LISTENER_EDGE = ProxyListenerEdge()
+# the ROUTER is part of the edge proxy. use the router to inject custom handlers that are handled before actual
+# service calls
+ROUTER: Router[Handler] = Router(
+    dispatcher=handler_dispatcher(), converters={"regex": RegexConverter}
+)
 
 
 def is_trace_logging_enabled(headers):
@@ -520,21 +543,25 @@ def is_trace_logging_enabled(headers):
 
 
 def do_start_edge(bind_address, port, use_ssl, asynchronous=False):
+    from localstack.http.adapters import RouterListener
     from localstack.services.internal import LocalstackResourceHandler
 
-    # add internal routes as default listener
-    ProxyListener.DEFAULT_LISTENERS.append(LocalstackResourceHandler())
-
     start_dns_server(asynchronous=True)
 
+    listeners = [
+        LocalstackResourceHandler(),  # handle internal resources first
+        RouterListener(ROUTER),  # then custom routes
+        PROXY_LISTENER_EDGE,  # then call the edge proxy listener
+    ]
+
     # get port and start Edge
     print("Starting edge router (http%s port %s)..." % ("s" if use_ssl else "", port))
-    # use use=True here because our proxy allows both, HTTP and HTTPS traffic
+    # use use_ssl=True here because our proxy allows both, HTTP and HTTPS traffic
     proxy = start_proxy_server(
         port,
         bind_address=bind_address,
         use_ssl=True,
-        update_listener=PROXY_LISTENER_EDGE,
+        update_listener=listeners,
         check_port=False,
     )
     if not asynchronous:
@@ -649,7 +676,8 @@ def run_process_as_sudo(component, port, asynchronous=False, env_vars=None):
         sudo_cmd,
         env_vars_str,
         python_cmd,
-        __file__,
+        "-m",
+        "localstack.services.edge",
         component,
         str(port),
     ]
