@@ -82,9 +82,14 @@ def dynamodbstreams():
 
 @aws_provider()
 def ec2():
-    from localstack.services.ec2 import ec2_listener, ec2_starter
+    from localstack.services.ec2.provider import Ec2Provider
+    from localstack.services.moto import MotoFallbackDispatcher
 
-    return Service("ec2", listener=ec2_listener.UPDATE_EC2, start=ec2_starter.start_ec2)
+    provider = Ec2Provider()
+    return Service(
+        "ec2",
+        listener=AwsApiListener("ec2", MotoFallbackDispatcher(provider)),
+    )
 
 
 @aws_provider()
@@ -152,11 +157,10 @@ def awslambda():
 
 @aws_provider()
 def logs():
-    from localstack.services.logs import logs_listener, logs_starter
+    from localstack.services.logs.provider import LogsAwsApiListener
 
-    return Service(
-        "logs", listener=logs_listener.UPDATE_LOGS, start=logs_starter.start_cloudwatch_logs
-    )
+    listener = LogsAwsApiListener()
+    return Service("logs", listener=listener)
 
 
 @aws_provider()
@@ -204,14 +208,25 @@ def s3():
 
 
 @aws_provider()
-def secretsmanager():
-    from localstack.services.secretsmanager import secretsmanager_listener, secretsmanager_starter
+def s3control():
+    from localstack.services.moto import MotoFallbackDispatcher
+    from localstack.services.s3control.provider import S3ControlProvider
 
+    provider = S3ControlProvider()
+    return Service(
+        "s3control", listener=AwsApiListener("s3control", MotoFallbackDispatcher(provider))
+    )
+
+
+@aws_provider()
+def secretsmanager():
+    from localstack.services.moto import MotoFallbackDispatcher
+    from localstack.services.secretsmanager.provider import SecretsmanagerProvider
+
+    provider = SecretsmanagerProvider()
     return Service(
         "secretsmanager",
-        listener=secretsmanager_listener.UPDATE_SECRETSMANAGER,
-        start=secretsmanager_starter.start_secretsmanager,
-        check=secretsmanager_starter.check_secretsmanager,
+        listener=AwsApiListener("secretsmanager", MotoFallbackDispatcher(provider)),
     )
 
 
@@ -298,20 +313,28 @@ def swf():
 
 @aws_provider()
 def resourcegroupstaggingapi():
-    from localstack.services.resourcegroupstaggingapi import rgta_listener, rgta_starter
+    from localstack.services.moto import MotoFallbackDispatcher
+    from localstack.services.resourcegroupstaggingapi.provider import (
+        ResourcegroupstaggingapiProvider,
+    )
 
+    provider = ResourcegroupstaggingapiProvider()
     return Service(
         "resourcegroupstaggingapi",
-        listener=rgta_listener.UPDATE_RGSA,
-        start=rgta_starter.start_rgsa,
+        listener=AwsApiListener("resourcegroupstaggingapi", MotoFallbackDispatcher(provider)),
     )
 
 
 @aws_provider(api="resource-groups")
 def resource_groups():
-    from localstack.services.resourcegroups import rg_listener, rg_starter
+    from localstack.services.moto import MotoFallbackDispatcher
+    from localstack.services.resourcegroups.provider import ResourceGroupsProvider
 
-    return Service("resource-groups", listener=rg_listener.UPDATE_RG, start=rg_starter.start_rg)
+    provider = ResourceGroupsProvider()
+    return Service(
+        "resource-groups",
+        listener=AwsApiListener("resource-groups", MotoFallbackDispatcher(provider)),
+    )
 
 
 @aws_provider()
